#ifndef _AMDCHIP_H_
#define	_AMDCHIP_H_

#include "amdchip.h"
#include "memory.h"

FlashCart *AtariMaxCart = NULL;

// From Cartridge.c
extern UBYTE *cart_image;
extern char cart_filename[FILENAME_MAX];

#define PAGE_LENGTH 256

#define FlashCartActiveBank (fc->activeChip->ActiveBank || (fc->activeChip==fc->Rom2) ? 0x40 : 0x00)

FlashCart *CreateFlashCart(UBYTE banks)
{
  FlashCart *fc = (FlashCart *)calloc(sizeof(FlashCart),1);
  fc->bModified  = FALSE;
  fc->Active     = TRUE;
  fc->TotalBanks = banks;
  switch (fc->TotalBanks)
  {
  case 16:
	  fc->Rom1 = CreateAmdChip(Am010, fc, 0);
	  break;
  case 64:
  case 128:
	  fc->Rom1 = CreateAmdChip(Am040, fc, 0);
	  break;
  default:
	  fc->Rom1 = CreateAmdChip(AmNone, fc, 0);
  }
  fc->activeChip = fc->Rom1;
  if (fc->TotalBanks == 128)
  {
	  fc->Rom2 = CreateAmdChip(Am040, fc, 1);
	  fc->activeChip = fc->Rom2;
  }
  else
  {
	  fc->Rom2 = CreateAmdChip(AmNone, fc, 0);
  }
  return fc;
}

void FreeFlashCart(FlashCart *fc)
{
	if (fc)
	{
		if (fc->bModified) FlashCartSave(fc, cart_filename);

		FreeAmdChip(fc->Rom1);
		FreeAmdChip(fc->Rom2);
		free(fc);
	}
}

void FlashCartFileRead(FlashCart *fc, FILE *fp)
{
	if (fc)
	{
		AmdChipFileRead(fc->Rom1, fp);
		AmdChipFileRead(fc->Rom2, fp);
		fc->bModified = FALSE;
	}
}

void FlashCartFileWrite(FlashCart *fc, FILE *fp)
{
	if (fc)
	{
		AmdChipFileWrite(fc->Rom1, fp);
		AmdChipFileWrite(fc->Rom2, fp);
		fc->bModified = FALSE;
	}
}

void FlashCartColdStart(FlashCart *fc)
{
	if (fc)
	{
		if (fc->bModified) FlashCartSave(fc, cart_filename);

		FlashCartAccessD5(fc, 0xD500+((fc->TotalBanks==16) ? 0 : fc->TotalBanks-1));
	}
}

int MapFlashCart(FlashCart *fc)
{
	if (fc && fc->Active)
	{
		CartA0BF_Enable();
		MapAmdChip(fc->activeChip);
#ifndef PAGED_ATTRIB
		SetHARDWARE(0xa000, 0xbfff);
#else
		for(i=0xA0;i<0xC0;i++) {
			readmap[i] = FlashCartReadHook;
			writemap[i] = FlashCartWriteHook;
		}
#endif
		return FlashCartActiveBank;
  }
  else
  {
	  CartA0BF_Disable();
  }
  // Cart is not active.
  return 0xFF;
}

void FlashCartSave(FlashCart *fc, const char *path)
{
	FILE *fp = fopen(path, "wb");
	 if (fp != NULL)
	 {
		 FlashCartFileWrite(fc, fp);
		 fclose(fp);
	 }
}

UBYTE FlashCartReadHook(UWORD mem)
{
	return FlashCartAreaRead(AtariMaxCart, mem);
}

UBYTE FlashCartAreaRead(FlashCart *fc, UWORD mem)
{
	return (fc) ? AmdChipRomRead(fc->activeChip,mem) : 0xFF;
}

void FlashCartWriteHook(UWORD mem,UBYTE val)
{
	AmdChipRomWrite(AtariMaxCart->activeChip, mem, val);
}

int FlashCartAccessD5(FlashCart *fc, UWORD mem)
{
	int bank = mem & 0xFF;

	if (fc)
	{
		//
		// Check whether the write goes to 0xd510. If so, deactivate the
		// cart.
		if ( ((fc->TotalBanks == 16) && (mem >= 0xd510 && mem <= 0xd51F)) ||
			 ((fc->TotalBanks == 128) && (mem == 0xd580)) )
		{
			if (fc->Active)
			{
				int ignore;
				// Deactivate the cart here.
				fc->Active = FALSE;
				fc->activeChip = NULL;
				ignore = MapFlashCart(fc);
			}
		}
		else if (mem <= (0xD500+fc->TotalBanks-1))
		{
			// Otherwise, select the proper bank.
			UBYTE newbank = (mem - 0xd500) % fc->TotalBanks;
			enum ChipType ct = GetChipType(fc->Rom2);
			AmdChip *newchip = fc->Rom1;
			if (ct != AmNone)
			{
				newchip = (newbank&0x40) ? fc->Rom2 : fc->Rom1;
				newbank &= 0x3F;
			}
			if (fc->activeChip != newchip || newbank != fc->activeChip->ActiveBank || fc->Active == FALSE)
			{
				// Ok, mapping changed, tell the mmu.
				fc->Active     = TRUE;
				fc->activeChip = newchip;
				fc->activeChip->ActiveBank = newbank;
				bank = MapFlashCart(fc);
			}
		}
	}

	return bank;
}

AmdChip *CreateAmdChip(enum ChipType ct, FlashCart *fc, int two_chips)
{
	AmdChip *chip = (AmdChip *)calloc(sizeof(AmdChip),1);
	chip->parent = fc;
	chip->TotalBanks = (ct == Am010) ? 0x10 : (ct == Am020) ? 0x20 : (ct == Am040) ? 0x40 : 0x00;
	chip->TotalPages = chip->TotalBanks << 5;
	if (ct != AmNone)
	{
		chip->Rom = cart_image + (PAGE_LENGTH * (two_chips ? fc->Rom1->TotalPages : 0));
	}
	else
		chip->Rom = NULL;
	chip->type = ct;
	chip->ActiveBank = (ct == Am010) ? 0 : chip->TotalBanks-1;
	chip->cmdState = CmdRead;
	return chip;
}

void FreeAmdChip(AmdChip *chip)
{
	if (chip)
	{
		chip->Rom = NULL;
		free(chip);
	}
}

enum ChipType GetChipType(AmdChip *chip)
{
	return (chip) ? chip->type : AmNone;
}

/// ChipErase
void AmdChipErase(AmdChip *chip, int sector)
{
	ULONG offset;
    int i, page_start=0, page_end=chip->TotalPages;
    if (sector != -1)
    {
        int sector_size; // pages
        switch (chip->type)
		{
        case Am010:
            sector_size = 0x40;
            page_start = sector_size * sector;
            page_end = page_start + sector_size - 1;
            break;
        case Am040:
            sector_size = 0x100;
            page_start = sector_size * sector;
            page_end = page_start + sector_size - 1;
            break;
        default:
            sector_size = 0x00;
            page_start = page_end = 0;
        }
     }

     for (i=page_start; i<page_end; i++)
	 {
		offset = PAGE_LENGTH*i;
		memset(chip->Rom+offset,255,PAGE_LENGTH);
	 }
}
///

/// RomAreaWrite
BOOL AmdChipRomWrite(AmdChip *chip, UWORD mem, UBYTE val)
{
	unsigned char *fp;
	UBYTE b;
	ULONG offset;
	UWORD m;
    // Writes to ROM are not allowed and are silently ignored
	m = (chip->ActiveBank & 7) << 13;
	m = m | (mem & 0x1FFF);
	if (chip->cmdState == CmdAutoSelect) chip->cmdState = CmdRead;
	switch (chip->cmdState)
	{
	case CmdRead:
		if (m == 0x5555 && val == 0xAA)
			chip->cmdState = CmdSecond;
		break;
	case CmdSecond:
		if (m == 0x2AAA && val == 0x55)
			chip->cmdState = CmdThird;
		break;
	case CmdThird:
		if (m == 0x5555)
		{
			switch (val)
			{
			case 0x90:
				chip->cmdState = CmdAutoSelect;
				break;
			case 0xA0:
				chip->cmdState = CmdProgram;
				break;
			case 0x80:
				chip->cmdState = CmdErase1;
				break;
			}
		}
		break;
	case CmdProgram:
		offset = (0x2000L*chip->ActiveBank)+(mem-0xA000);
		fp = chip->Rom+offset;
		b = *fp;
		*fp &= val;
		chip->cmdState = CmdRead;
		chip->parent->bModified = TRUE;
		break;
	case CmdErase1:
		if (m == 0x5555 && val == 0xAA)
			chip->cmdState = CmdErase2;
		break;
	case CmdErase2:
		if (m == 0x2AAA && val == 0x55)
			chip->cmdState = CmdErase;
		break;
	case CmdErase:
		if (m == 0x5555 && val == 0x10)
		{
			AmdChipErase(chip, -1);
			chip->parent->bModified = TRUE;
			chip->cmdState = CmdRead;
		}
		if (val == 0x30)
		{
            int sector;
            if (chip->type == Am010) sector = ((chip->ActiveBank&1)<<2) + ((mem>>13)& 3); // 16K
            if (chip->type == Am040) sector = ((chip->ActiveBank>>4)& 7); // 64K
			AmdChipErase(chip, sector);
			chip->parent->bModified = TRUE;
			chip->cmdState = CmdRead;
        }
		break;
	default:
		if (val == 0xF0) chip->cmdState = CmdRead;
	}
	return FALSE;
}

/// FlashCart::RomAreaRead
UBYTE AmdChipRomRead(AmdChip *chip, UWORD mem)
{
	UBYTE val = 0xFF;
	if (chip)
	{
		ULONG offset;
		UBYTE *p;
		offset = (0x2000L*chip->ActiveBank)+(mem-0xA000);
		p = chip->Rom+offset;
		val = *p;
		if (chip->cmdState == CmdAutoSelect)
		{
			switch (mem & 0xFF)
			{
			case 0:
				val = (chip->type != AmNone) ? 0x01 : 0x00; // AMD
				break;
			case 1:
				switch (chip->type)
				{
				 case Am010:
					  val = 0x20;
					  break;
				 case Am040:
					  val = 0xA4;
					  break;
				 default:
					  val = 0x00;
				 }
			}
		}
	}
	return val;
}

void AmdChipFileRead(AmdChip *chip, FILE *fp)
{
	ULONG offset;
	int i;
	for (i=0; i<chip->TotalPages; i++)
	{
		offset = i*PAGE_LENGTH;
    	memset(chip->Rom+offset,255,PAGE_LENGTH);
	    fread(chip->Rom+offset,PAGE_LENGTH,1,fp);
	}
}

void AmdChipFileWrite(AmdChip *chip, FILE *fp)
{
	ULONG offset;
	int i;
	for (i=0; i<chip->TotalPages; i++)
	{
		offset = i*PAGE_LENGTH;
	    fwrite(chip->Rom+offset,PAGE_LENGTH,1,fp);
	}
}

BOOL MapAmdChip(AmdChip *chip)
{
	if (chip->type != AmNone)
	{
		ULONG displacement;
		// Get the active bank and map it into 0xa000 to 0xbfff
		// 8K = 2^13
		displacement = (0x2000L*chip->ActiveBank);
		CopyROM(0xa000, 0xbfff, chip->Rom+displacement);
		return TRUE;
	}
	return FALSE;
}

#endif
